"""Reading and writing data."""

import numpy as np
import pandas as pd
import networkx as nx
from tqdm import tqdm
from scipy import sparse
#from texttable import Texttable
from sklearn import preprocessing

def read_graph(data):
    """
    Method to read graph and create a target matrix with pooled adjacency matrix powers.
    :param args: Arguments object.
    :return graph: graph.
    """
    print("\nTarget matrix creation started.\n")
    edges = data.values.tolist()
    edges = [[int(edge[0]), int(edge[1])] for edge in edges]
    graph = nx.from_edgelist(edges)
    graph.remove_edges_from(nx.selfloop_edges(graph))
    return graph



# def tab_printer(args):
#     """
#     Function to print the logs in a nice tabular format.
#     :param args: Parameters used for the model.
#     """
#     args = vars(args)
#     keys = sorted(args.keys())
#     t = Texttable()
#     t.add_rows([["Parameter", "Value"]])
#     t.add_rows([[k.replace("_", " ").capitalize(), args[k]] for k in keys])
#     print(t.draw())

def feature_calculator( graph,window_size):
    """
    Calculating the feature tensor.
    :param args: Arguments object.
    :param graph: NetworkX graph.
    :return target_matrices: Target tensor.
    """
    index_1 = [edge[0] for edge in graph.edges()] + [edge[1] for edge in graph.edges()]
    index_2 = [edge[1] for edge in graph.edges()] + [edge[0] for edge in graph.edges()]
    values = [1 for edge in index_1]
    node_count = max(len(set(index_1)), len(set(index_2)))
    adjacency_matrix = sparse.coo_matrix((values, (index_1, index_2)),
                                         shape=(node_count, node_count),
                                         dtype=np.float32)

    degrees = adjacency_matrix.sum(axis=0)[0].tolist()
    degs = sparse.diags(degrees, [0])
    normalized_adjacency_matrix = degs.dot(adjacency_matrix)
    target_matrices = [normalized_adjacency_matrix.todense()]
    powered_A = normalized_adjacency_matrix
    if window_size > 1:
        for power in tqdm(range(window_size-1), desc="Adjacency matrix powers"):
            powered_A = powered_A.dot(normalized_adjacency_matrix)
            to_add = powered_A.todense()
            target_matrices.append(to_add)
    target_matrices = np.array(target_matrices)
    #print(target_matrices)
    return target_matrices

def adjacency_opposite_calculator(graph):
    """
    Creating no edge indicator matrix.
    :param graph: NetworkX object.
    :return adjacency_matrix_opposite: Indicator matrix.
    """
    adjacency_matrix = sparse.csr_matrix(nx.adjacency_matrix(graph), dtype=np.float32).todense()
    adjacency_matrix_opposite = np.ones(adjacency_matrix.shape) - adjacency_matrix
    return adjacency_matrix_opposite



import dgl
import dgl.function as fn
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from models.utils import *
from dgl.nn.pytorch import GraphConv
import time


gcn_msg = fn.copy_src(src='h', out='m')
gcn_reduce = fn.sum(msg='m', out='h')

#
# class GCNLayer(nn.Module):
#     def __init__(self, in_feats, out_feats):
#         super(GCNLayer, self).__init__()
#         self.linear = nn.Linear(in_feats, out_feats)
#
#     def forward(self, g, feature):
#         # Creating a local scope so that all the stored ndata and edata
#         # (such as the `'h'` ndata below) are automatically popped out
#         # when the scope exits.
#         with g.local_scope():
#             g.ndata['h'] = feature
#             g.update_all(gcn_msg, gcn_reduce)
#             h = g.ndata['h']
#             return self.linear(h)

class Net(nn.Module):
    def __init__(self,in_feats, out_feats,num_classes):
        super(Net, self).__init__()
        self.layer1 = GraphConv(in_feats, out_feats)
        self.layer2 = GraphConv(out_feats, num_classes)

    def forward(self, g, features):
        h = self.layer1(g, features)
        h = torch.relu(h)
        h = self.layer2(g, h)
        return h




class GCN(object):
    def __init__(self, data, embed_size = 128, task = 'NodeClassification',train_test_ratio = 0.8):
        self.g,self.features,self.labels,self.train_mask,self.test_mask = build_graph(data, train_test_ratio,embed_size=128)
        self.embed_size = embed_size

    def evaluate(self,model, g, features, labels, mask):
        model.eval()
        with th.no_grad():
            logits = model(g, features)
            logits = logits[mask]
            labels = labels[mask]
            _, indices = th.max(logits, dim=1)
            correct = th.sum(indices == labels)
            return correct.item() * 1.0 / len(labels)

    def train(self,verbose = 1,epochs = 10):
        net = Net(self.g.number_of_nodes(), self.embed_size,num_classes=len(th.unique(self.labels)))
        print(net)
        optimizer = th.optim.Adam(net.parameters(), lr=1e-2)
        dur = []
        for epoch in range(epochs):
            if epoch >=3:
                t0 = time.time()

            net.train()
            logits = net(self.g, self.features)
            logp = F.log_softmax(logits, 1)
            loss = F.nll_loss(logp[self.train_mask], labels[self.train_mask])

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if epoch >=3:
                dur.append(time.time() - t0)

            acc = evaluate(net, self.g, self.features, self.labels, self.test_mask)
            print("Epoch {:05d} | Loss {:.4f} | Test Acc {:.4f} | Time(s) {:.4f}".format(
                    epoch, loss.item(), acc, np.mean(dur)))



'''
input data:
a dataframe with two columns: user item

output:
a embedding lookup dictionary {'user_id/item_id':[vector]}

'''
from gensim.models import Word2Vec
#import deepneighbor.config as config
from deepneighbor.utils import generate_sentences,generate_sentences_dw,convert_to
from annoy import AnnoyIndex
from sklearn import preprocessing
from deepneighbor.models.gat import *



class Embed(object):
    def __init__(self,data_path,model='w2v',num_walks=100,walk_length=10):
        '''
        data: a dataframe: user, item

        model: 'w2v','deepwalk','gcn', 'gat'
        '''

        le = preprocessing.LabelEncoder()
        self.data ,self.le = convert_to(data_path,le)
        self.w2v_model = None
        self._annoy = None

        self._embeddings = dict()
        self.model_type = model
        self.num_walks = num_walks
        self.walk_length = walk_length

        if self.model_type == 'w2v':

            self.graph, self.sentences = generate_sentences(self.data)

        if self.model_type == 'deepwalk':
            self.sentences = generate_sentences_dw(data)
        if self.model_type == 'gat':
            pass




    def train(self,
            window_size=5,
            workers=3,
            iter=5,
            learning_rate=0.01,
            epochs = 10,
            dimensions = 128,
            num_of_walks=80,
            beta=0.5,
            gamma=0.5,
            **kwargs):
        self.workers=workers
        self.iter=iter
        self.window_size=window_size
        self.learning_rate=learning_rate
        self.epochs=epochs
        self.dimensions=dimensions
        self.num_of_walks=num_of_walks
        self.beta=beta
        self.gamma=gamma
        self._annoy = AnnoyIndex(dimensions, 'angular')

        if self.model_type == 'w2v' or self.model_type=='deepwalk':
            kwargs["sentences"] = self.sentences
            kwargs["min_count"] = kwargs.get("min_count", 0)
            kwargs["size"] = self.dimensions
            kwargs["sg"] = 1  # skip gram
            kwargs["hs"] = 1  # deepwalk use Hierarchical Softmax
            kwargs["workers"] = self.workers
            kwargs["window"] = self.window_size
            kwargs["iter"] = self.iter


            print(f"There are {self.data.user.nunique()} users")
            print(f"There are {self.data.item.nunique()} items")

            print("Learning embedding vectors...")
            model_w2v = Word2Vec(**kwargs)
            print("Learning embedding vectors done!")

            self.w2v_model = model_w2v



            #words = self.data['user'].unique().tolist() + self.data['item'].unique().tolist()

            for node in self.w2v_model.wv.vocab:
                # node = "'" + str(node) +"'"
                #print(node)
                #print(self.w2v_model.wv[node])
                self._annoy.add_item(int(node),self.w2v_model.wv[node])
                #self.


            self._annoy.build(-1)

        if self.model_type == 'gat':
            model = AttentionWalkTrainer(graph_path=self.data,
                                        dimensions=self.dimensions ,
                                        learning_rate=self.learning_rate,
                                        epochs=self.epochs ,
                                        window_size=self.window_size ,
                                        num_of_walks=self.num_of_walks,
                                        beta=self.beta,
                                        gamma=self.gamma )
            model.fit()
            self.emb = model.save_embedding()

            for id in self.emb.id:
                self._annoy.add_item(int(id),self.emb[self.emb.id==id].values.tolist()[0][1:])

            self._annoy.build(-1)




        #return model_w2v


    def get_embeddings(self,):
        if self.model_type == 'w2v' or self.model_type == 'deepwalk':
            for node in self.w2v_model.wv.vocab:
                #print(node)
                #print(self.le.inverse_transform([int(node)]))
                self._embeddings[self.le.inverse_transform([int(node)])[0]] = self.w2v_model.wv[node]

        if self.model_type == 'gat':

            for id in self.emb.id:
                self._embeddings[self.le.inverse_transform([int(id)])[0]]= self.emb[self.emb.id==id].values.tolist()[0][1:]

        return self._embeddings

    def search(self, seed,k = 5, type=None):
        '''
        seed: seed item to find nearest neighbor
        k: number of cloest neighhbors
        '''

        a_return = self._annoy.get_nns_by_item(int(self.le.transform([seed])[0]), k)
        return list(self.le.inverse_transform(a_return))



import pandas as pd

if __name__ == "__main__":
    data = pd.read_csv("datasets/data.csv",nrows=10000)
    data = data[['item_id','user_id']]
    data.columns = ['item','user']
    data['item'] = data['item'].astype(str)
    data['user'] = data['user'].astype(str)
    data.to_csv("datasets/processed_data.csv",index=None)
    #print(data.head())
    #print("read 100 rows!")



import pandas as pd
import embed
from embed import Embed

if __name__ == "__main__":
    data_path = 'datasets/processed_data.csv'
    emb = Embed(data_path,model = 'gat',)
    emb.train(epochs=1)
    print("Trained")
    #print(emb.sentences)
    #emb.get_embeddings().shape
    print(emb.search('7443',k=10))
    print(len(emb.get_embeddings()['7443']))

    print("Found!")
    #return embed.get_embeddings()



import networkx as nx
import itertools
import random
import pandas as pd


def generate_sentences(data):
    '''
    input dataframe.
      user  item
    	1	a
    	1	b
    	1	c
    	2	d
    	2	a
    output sentences. [user_id1, all_item(1...k1),user_id2,all_item(1...k2),...]
    [1,a,b,c,2,d,a]
    '''
    sentences = []

    edges = data.values.tolist()
    edges = [[int(edge[0]), int(edge[1])] for edge in edges]
    graph = nx.from_edgelist(edges)
    graph.remove_edges_from(nx.selfloop_edges(graph))

    for node in graph.nodes():
        temp = []
        temp = [str(node)] + [str(n) for n in list(graph.neighbors(node))]
        sentences.append(temp)
    return graph, sentences

def convert_to(data,le):
    '''
    given txt or csv file
    transform with label encoder
    return dataframe with two columns [from] and [to]
    '''

    data = pd.read_csv(data)
    data.columns = ['user','item']
    le.fit(data.user.unique().tolist()+data.item.unique().tolist())
    for col in data.columns:
        data[col] = le.transform(data[col])
    return data,le




def generate_sentences_dw(data):
    return nx.read_edgelist(data,)



import argparse

def parameter_parser():
    """
    A method to parse up command line parameters.
    By default it gives an embedding of the Wikipedia Chameleons dataset.
    The default hyperparameters give a good quality representation without grid search.
    Representations are sorted by node ID.
    """
    parser = argparse.ArgumentParser(description="Run Attention Walk.")

    parser.add_argument("--edge-path",
                        nargs="?",
                        default="./datasets/chameleon_edges.csv",
	                help="Edge list csv.")

    parser.add_argument("--embedding-path",
                        nargs="?",
                        default="./output/chameleon_AW_embedding.csv",
	                help="Target embedding csv.")

    parser.add_argument("--attention-path",
                        nargs="?",
                        default="./output/chameleon_AW_attention.csv",
	                help="Attention vector csv.")

    parser.add_argument("--dimensions",
                        type=int,
                        default=128,
	                help="Number of dimensions. Default is 128.")

    parser.add_argument("--epochs",
                        type=int,
                        default=200,
	                help="Number of gradient descent iterations. Default is 200.")

    parser.add_argument("--window-size",
                        type=int,
                        default=5,
	                help="Skip-gram window size. Default is 5.")

    parser.add_argument("--num-of-walks",
                        type=int,
                        default=80,
	                help="Number of random walks. Default is 80.")

    parser.add_argument("--beta",
                        type=float,
                        default=0.5,
	                help="Regularization parameter. Default is 0.5.")

    parser.add_argument("--gamma",
                        type=float,
                        default=0.5,
	                help="Regularization parameter. Default is 0.5.")

    parser.add_argument("--learning-rate",
                        type=float,
                        default=0.01,
	                help="Gradient descent learning rate. Default is 0.01.")

    return parser.parse_args()



import torch
import numpy as np
import pandas as pd
from tqdm import trange
from models.utils import read_graph, feature_calculator, adjacency_opposite_calculator
from sklearn import preprocessing

class AttentionWalkLayer(torch.nn.Module):
    """
    Attention Walk Layer.
    For details see the paper.
    """
    def __init__(self,graph_path,window_size,learning_rate,epochs,dimensions,num_of_walks,beta,gamma,shapes):
        """
        Setting up the layer.
        :param args: Arguments object.
        :param shapes: Shape of the target tensor.
        """
        super(AttentionWalkLayer, self).__init__()
        self.graph_path = graph_path
        self.window_size = window_size
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.num_of_walks = num_of_walks
        self.dimensions = dimensions
        self.shapes = shapes
        self.beta=beta
        self.gamma = gamma
        self.define_weights()
        self.initialize_weights()

    def define_weights(self):
        """
        Define the model weights.
        """
        half_dim = int(self.dimensions/2)
        self.left_factors = torch.nn.Parameter(torch.Tensor(self.shapes[1], half_dim))
        self.right_factors = torch.nn.Parameter(torch.Tensor(half_dim, self.shapes[1]))
        self.attention = torch.nn.Parameter(torch.Tensor(self.shapes[0], 1))

    def initialize_weights(self):
        """
        Initializing the weights.
        """
        torch.nn.init.uniform_(self.left_factors, -0.01, 0.01)
        torch.nn.init.uniform_(self.right_factors, -0.01, 0.01)
        torch.nn.init.uniform_(self.attention, -0.01, 0.01)

    def forward(self, weighted_target_tensor, adjacency_opposite):
        """
        Doing a forward propagation pass.
        :param weighted_target_tensor: Target tensor factorized.
        :param adjacency_opposite: No-edge indicator matrix.
        :return loss: Loss being minimized.
        """
        self.attention_probs = torch.nn.functional.softmax(self.attention, dim=0)
        probs = self.attention_probs.unsqueeze(1).expand_as(weighted_target_tensor)
        weighted_target_tensor = weighted_target_tensor * probs
        weighted_tar_mat = torch.sum(weighted_target_tensor, dim=0)
        weighted_tar_mat = weighted_tar_mat.view(self.shapes[1], self.shapes[2])
        estimate = torch.mm(self.left_factors, self.right_factors)
        loss_on_target = - weighted_tar_mat* torch.log(torch.sigmoid(estimate))
        loss_opposite = -adjacency_opposite * torch.log(1-torch.sigmoid(estimate))
        loss_on_mat = self.num_of_walks*weighted_tar_mat.shape[0]*loss_on_target+loss_opposite
        abs_loss_on_mat = torch.abs(loss_on_mat)
        average_loss_on_mat = torch.mean(abs_loss_on_mat)
        norms = torch.mean(torch.abs(self.left_factors))+torch.mean(torch.abs(self.right_factors))
        loss_on_regularization = self.beta * (self.attention.norm(2)**2)
        loss = average_loss_on_mat + loss_on_regularization + self.gamma*norms
        return loss

class AttentionWalkTrainer(object):
    """
    Class for training the AttentionWalk model.
    """
    def __init__(self,
                graph_path,
                window_size,
                learning_rate,
                epochs ,
                dimensions ,
                num_of_walks,

                beta,
                gamma):
        """
        Initializing the training object.
        :param args: Arguments object.
        """
        self.graph_path = graph_path
        self.window_size = window_size
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.dimensions = dimensions
        self.beta = beta
        self.gamma = gamma
        self.num_of_walks=num_of_walks
        #self.le = preprocessing.LabelEncoder()

        self.graph = read_graph(self.graph_path)
        self.initialize_model_and_features()

    def initialize_model_and_features(self):
        """
        Creating data tensors and factroization model.
        """
        print("Initialing Graph data.")
        self.target_tensor = feature_calculator( self.graph,self.window_size)

        self.target_tensor = torch.FloatTensor(self.target_tensor)

        self.adjacency_opposite = adjacency_opposite_calculator(self.graph)
        self.adjacency_opposite = torch.FloatTensor(self.adjacency_opposite)
        print(f"{self.target_tensor.shape} and {self.adjacency_opposite.shape}")
        self.model = AttentionWalkLayer(graph_path = self.graph_path,
                                        window_size = self.window_size,
                                        learning_rate = self.learning_rate,
                                        epochs = self.epochs,
                                        num_of_walks=self.num_of_walks,
                                        dimensions = self.dimensions,
                                        beta = self.beta,
                                        gamma = self.gamma,
                                        shapes = self.target_tensor.shape)




    def fit(self):
        """
        Fitting the model
        """
        print("\nTraining the model.\n")
        self.model.train()
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate)
        self.epochs = trange(self.epochs, desc="Loss")
        for _ in self.epochs:
            self.optimizer.zero_grad()
            loss = self.model(self.target_tensor, self.adjacency_opposite)
            loss.backward()
            self.optimizer.step()
            self.epochs.set_description("Attention Walk (Loss=%g)" % round(loss.item(), 4))
        print("\nModel Training Finished.\n")

    def save_model(self):
        """
        Saving the embedding and attention vector.
        """
        self.save_embedding()
        self.save_attention()

    def save_embedding(self):
        """
        Saving the embedding matrices as one unified embedding.
        """
        print("\nSaving the model.\n")
        left = self.model.left_factors.detach().numpy()
        right = self.model.right_factors.detach().numpy().T
        indices = np.array([range(len(self.graph))]).reshape(-1, 1)
        embedding = np.concatenate([indices, left, right], axis=1)
        columns = ["id"] + ["x_" + str(x) for x in range(self.dimensions)]
        embedding = pd.DataFrame(embedding, columns=columns)
        #embedding.to_csv(self.args.embedding_path, index=None)
        print("\nEmbedding Saved!.\n")
        return embedding

    def save_attention(self):
        """
        Saving the attention vector.
        """
        attention = self.model.attention_probs.detach().numpy()
        indices = np.array([range(self.window_size)]).reshape(-1, 1)
        attention = np.concatenate([indices, attention], axis=1)
        attention = pd.DataFrame(attention, columns=["Order", "Weight"])
        #attention.to_csv(self.args.attention_path, index=None)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
REQUIRED_PACKAGES = [
    'h5py',
    'requests',
    'gensim==3.7.0',
    'tqdm',
    'numpy',
    'scikit-learn',
    'pandas',
    'annoy',
    'dgl',
    'torch>=1.1.0',
    'tqdm'

]

setuptools.setup(

  name="deepneighbor",
  version="0.3.0",
  author="Yufeng Wang",
  author_email="louiswang524@gmail.com",
  description="embedding-based item nearest neighborhoods extraction",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/LouisBIGDATA/deepneighbor",
  packages=setuptools.find_packages(),
  python_requires=">=3.4",  # '>=3.4',  # 3.4.6
    install_requires=REQUIRED_PACKAGES,
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  'Intended Audience :: Developers',
 'Intended Audience :: Education',
 'Intended Audience :: Science/Research',
  'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
  ],
  keywords=['embedding', 'information retrieval',
              'deep learning', 'torch', 'tensor', 'pytorch','nearest neighbor']
)
